﻿
Module ThereminoSystem



    Friend WithEvents _NetHALRPC As NetHALRPC = New NetHALRPC
    Private ProgressiveModuleIdCounter As Int32

    Friend DuplicateNames As Boolean = False
    Friend ConfigurationIsValid As Boolean = False
    Friend InOutModules As List(Of Module_InOut)
    Friend FreeRunningTimer As Stopwatch = New Stopwatch
    Friend CommandSlot As Int32 = 0

    ' ==================================================================================
    '  NetHALRPC REPLAYS
    ' ==================================================================================
    Private Sub ScanModuleReplay(ByVal ModuleProps As NetHALRPC.ScanReplayModulePropsStr, _
                                 ByVal PinsCaps() As String) Handles _NetHALRPC.ScanModuleReplay

        ' The SyncLock postpones a new entry if the function is not finished
        Static obj As Object = New Object
        SyncLock obj
            ' ------------------------------------------------------ if IP already exists then exit
            If FindModuleByIP(ModuleProps.IPAddr) IsNot Nothing Then Return
            ' ------------------------------------------------------ If there are names then test the Module-name
            With CommOptions
                If .Names.Length > 0 Then
                    If Array.IndexOf(.Names, ModuleProps.Name) < 0 Then
                        Return
                    End If
                End If
            End With
            ' ------------------------------------------------------ Increase IotModules Array and add new IP
            Dim m As Module_InOut
            m = New Module_InOut(ProgressiveModuleIdCounter)
            ProgressiveModuleIdCounter += 1
            m.IpAddr = ModuleProps.IPAddr
            m.SetName(ModuleProps.Name)
            m.ProgStatus = ModuleProps.ProgStatus
            m.ModuleVersion = "Ver:" + ModuleProps.FWVer + " IP:" + m.IpAddr
            m.MyDeviceIsActive = True
            m.MyDeviceIsProgrammed = False
            m.Set_Adc24_MasterPinIndex(PinsCaps)
            m.SetPinCaps(PinsCaps)
            InOutModules.Add(m)
        End SyncLock
    End Sub

    Private Sub ProgModuleReplay(ByVal ModuleProps As NetHALRPC.ProgModuleReplayPropsStr) Handles _NetHALRPC.ProgModuleReplay
        IotModulesArray_UpdateData(ModuleProps.IPAddr, _
                                   "", _
                                   ModuleProps.ProgStatus, _
                                   "")
        Dim m As Module_InOut = FindModuleByIP(ModuleProps.IPAddr)
        m.ProgModuleCompleted = True
        If ModuleProps.ProgStatus = "Y" Then
            m.MyDeviceIsProgrammed = True
        Else
            m.MyDeviceIsProgrammed = False
        End If
    End Sub

    Private Sub NameModuleReplay(ByVal ModuleName As NetHALRPC.NameModulePropsStr) Handles _NetHALRPC.NameModuleReplay
        IotModulesArray_UpdateData(ModuleName.IPAddr, _
                                   ModuleName.Name, _
                                   "", _
                                   "")
    End Sub

    Private Sub IotModulesArray_UpdateData(ByVal IP As String, _
                                           ByVal Name As String, _
                                           ByVal ProgStatus As String, _
                                           ByVal FWVer As String)
        For i As Int32 = 0 To InOutModules.Count - 1
            With InOutModules(i)
                If IP = .IpAddr Then
                    If Name <> "" Then .SetName(Name)
                    If ProgStatus <> "" Then .ProgStatus = ProgStatus
                    If FWVer <> "" Then .ModuleVersion = FWVer
                    Exit For
                End If
            End With
        Next
    End Sub


    ' ''TODO - Adc16 input = 0 --------------------

    Private Sub XchgModuleReplay(ByVal ModuleProps As NetHALRPC.XchgModuleReplayStr, ByVal PinsIn() As Single) Handles _NetHALRPC.XchgSingleModuleReplay
        Dim m As Module_InOut = FindModuleByIP(ModuleProps.IPAddr)
        m.ProgrammedStatus = ModuleProps.ProgStatus
        m.DataExchangeCompleted = True
        Dim i As Int32 = 0
        For Each PinValue As Single In PinsIn
            m.InputDataBuffer(i) = PinValue
            i += 1
        Next
    End Sub

    Private Sub GetSinglePinReplay(ByVal ModuleProps As String, ByVal PinsProps As NetHALRPC.GetSinglePinReplayStr) Handles _NetHALRPC.GetSinglePinReplay
        Dim m As Module_InOut = FindModuleByIP(ModuleProps)
        m.SinglePinPropsFromIotModule = PinsProps
        m.SinglePinPropsRequestCompleted = True
    End Sub

    Private Sub PwmPinDataReplay(ByVal DataReceived As NetHALRPC.DataReceived) Handles _NetHALRPC.PwmPinDataReplay
        Dim m As Module_InOut = FindModuleByIP(DataReceived.IPAddr)
        m.PwmPinDataCompleted = True
    End Sub

    Private Sub StepperPinDataReplay(ByVal DataReceived As NetHALRPC.DataReceived) Handles _NetHALRPC.StepperPinDataReplay
        Dim m As Module_InOut = FindModuleByIP(DataReceived.IPAddr)
        m.StepperPinDataOK = DataReceived.ReceivedOK
        m.StepperPinDataCompleted = True
    End Sub

    Private Sub Adc24ConfigDataReplay(ByVal DataReceived As NetHALRPC.DataReceived) Handles _NetHALRPC.Adc24ConfigDataReplay
        Dim m As Module_InOut = FindModuleByIP(DataReceived.IPAddr)
        m.Adc24ConfigCompleted = True
    End Sub

    ' ==================================================================================
    '  INIT MODULES
    ' ==================================================================================
    Friend Sub TheSystem_InitModules()
        Form1.Cursor = Cursors.WaitCursor
        ' ------------------------------------------------ used for timers with "convert to frequency"
        FreeRunningTimer.Start()
        ' -------------------------------------------------------------- clear module list
        If InOutModules IsNot Nothing Then
            For Each m As Module_InOut In InOutModules
                m.CloseCommunication()
            Next
            InOutModules.Clear()
        Else
            InOutModules = New List(Of Module_InOut)
        End If
        _NetHALRPC.CloseAdapters()
        ' --------------------------------------------------------------
        With CommOptions
            ' ---------------------------------------------------------- test for empty arrays
            If .Names Is Nothing OrElse .Names.Length = 0 OrElse .Names(0) = "all" Then
                ReDim .Names(-1)
            End If
            If .LanIpList Is Nothing OrElse .LanIpList.Length = 0 OrElse .LanIpList(0) = "all" Then
                ReDim .LanIpList(-1)
            End If
            ' ---------------------------------------------------------- open LAN Adapters
            _NetHALRPC.ScanAdapters(.LanPort)
        End With
        ' -------------------------------------------------------------- scan local area
        ProgressiveModuleIdCounter = 0
        For i As Int32 = 1 To 5
            _NetHALRPC.ScanLocalModuleRequest()
            SleepMyThread(10)
        Next
        ' -------------------------------------------------------------- scan wide area
        Dim timeout As Int32 = 100
        Dim scanrem As NetHALRPC.ScanRemoteModuleRequestStr
        For j As Int32 = 1 To 5
            For i As Int32 = 0 To CommOptions.IpOptionList.Length - 1
                With CommOptions.IpOptionList(i)
                    scanrem.IPAddr = .IP
                    scanrem.Port = .Port
                    If .Timeout > timeout Then timeout = .Timeout
                    _NetHALRPC.ScanRemoteModuleRequest(scanrem)
                End With
            Next
            SleepMyThread(timeout)
        Next
        ' -------------------------------------------------------------- Sort Modules
        InOutModules.Sort(AddressOf ModuleComparer)
        ' -------------------------------------------------------------- and reorder ID
        SetAllModuleId()
        ' -------------------------------------------------------------- 
        Form1.Cursor = Cursors.Default
    End Sub

    Private Function ModuleComparer(ByVal m1 As Module_InOut, ByVal m2 As Module_InOut) As Integer
        Return String.Compare(m1.GetName(), m2.GetName(), True)
    End Function

    Friend Sub TheSystem_DisconnectModule(ByVal m As Module_InOut)
        If m Is Nothing Then Return
        m.CloseCommunication()
        InOutModules.Remove(m)
        SetAllModuleId()
    End Sub

    Private Sub SetAllModuleId()
        For i As Int32 = 0 To InOutModules.Count - 1
            InOutModules(i).ModuleId = i
            If InOutModules(i).Pins IsNot Nothing Then
                For Each p As Pin In InOutModules(i).Pins
                    p.ModuleId = i
                Next
            End If
        Next
    End Sub

    Friend Sub TheSystem_StartTimers()
        If Not ConfigurationIsValid Then Exit Sub
        For Each m As Module_InOut In InOutModules
            m.StartTimedOperations()
        Next
    End Sub

    Friend Sub TheSystem_StopTimers()
        For Each m As Module_InOut In InOutModules
            m.StopTimedOperations()
        Next
    End Sub

    Friend Sub TheSystem_AssignConfigurationToModules()
        ' ------------------------------------------------ assign configuration
        AssignDefaultSlotsToPins()
        AssignConfigurationsToModules()
        SetConfigurationParamsToDevices()
        ' ------------------------------------------------ show pins in the list
        Form1.SelectedLine = -1
        TheSystem_ListComponents()
        Form1.ListView_SetAllLineColors()
        MarkErrorLines()
        Form1.ToolStripButton_Validate.Enabled = Not ConfigurationIsValid And Not DuplicateNames
        Form1.EnableCalibrateButton(TheSystem_CalibrationNeeded)
        TheSystem_Calibrate()
    End Sub

    Private Sub AssignDefaultSlotsToPins()
        Dim i As Int32 = 1
        For Each m As Module_InOut In InOutModules
            Dim j As Int32 = i
            i += 100
            For Each p As Pin In m.Pins
                p.Slot = j
                j += 1
            Next
        Next
    End Sub

    Friend Function TheSystem_CalibrationNeeded() As Boolean
        For Each m As Module_InOut In InOutModules
            If m.CalibrationNeeded Then Return True
        Next
    End Function

    Friend Sub TheSystem_Calibrate()
        For Each m As Module_InOut In InOutModules
            m.Calibrate()
        Next
    End Sub

    Friend Sub TheSystem_CalibrateZero()
        For Each m As Module_InOut In InOutModules
            m.CalibrateZero()
        Next
    End Sub

    Friend Sub TheSystem_SetParkingPositions()
        For Each m As Module_InOut In InOutModules
            m.SetOutputPinsToParkingPosition()
        Next
    End Sub


    ' ==================================================================================================
    '   FIND COMPONENTS FROM LIST
    ' ==================================================================================================
    Friend Function FindModuleByListLine(ByVal line As Int32) As Module_InOut
        If InOutModules Is Nothing Then Return Nothing
        Dim n As Int32 = 0
        For Each m As Module_InOut In InOutModules
            If n = line Then Return m
            n += 1
            For Each p As Pin In m.Pins
                If n = line Then Return m
                n += 1
            Next
        Next
        Return Nothing
    End Function

    Private Function FindModuleByIP(ByVal ip As String) As Module_InOut
        For Each m As Module_InOut In InOutModules
            If m.IpAddr = ip Then
                Return m
            End If
        Next
        Return Nothing
    End Function

    Friend Function FindPinByListLine(ByVal line As Int32) As Pin
        Dim n As Int32 = 0
        For Each m As Module_InOut In InOutModules
            n += 1
            For Each p As Pin In m.Pins
                If n = line Then Return p
                n += 1
            Next
        Next
        Return Nothing
    End Function

    Friend Function GetSelectedPin() As Pin
        Return FindPinByListLine(Form1.SelectedLine)
    End Function

    Friend Function GetSelectedListLine() As Int32
        Return Form1.SelectedLine
    End Function


    ' ==================================================================================================
    '   LIST COMPONENTS
    ' ==================================================================================================
    Friend Sub TheSystem_ListComponents()

        LoadSlotNames()

        Dim item As ListViewItem
        With Form1.MyListView1
            .OwnerDraw = True
            .GridLines = True
            .FullRowSelect = True
            .View = View.Details
            .LabelEdit = False
            .Font = New Font("Courier new", 8)
            ' ------------------------------------------------------------------ column headers
            .Header_BackColor1 = Color.FromArgb(255, 255, 200)
            .Header_BackColor2 = Color.FromArgb(200, 200, 200)
            .Header_ForeColor = Color.FromArgb(20, 20, 20)
            .Header_ShadowColor = Color.FromArgb(0, 255, 255, 200)
            .Header_Font = New Font("Microsoft Sans Serif", 9, FontStyle.Regular)
            '
            ' -----------------------------------------------------------------
            .Clear()
            .Header_TextVerticalPosition = -1
            .Header_TextAlignments = New StringAlignment() {StringAlignment.Near, _
                                                            StringAlignment.Center, _
                                                            StringAlignment.Near, _
                                                            StringAlignment.Center, _
                                                            StringAlignment.Center, _
                                                            StringAlignment.Near}

            .Columns.Add(" Type", 56, HorizontalAlignment.Left)
            .Columns.Add("ID", 28, HorizontalAlignment.Left)
            .Columns.Add(" Subtype", 105, HorizontalAlignment.Left)
            .Columns.Add("Slot", 35, HorizontalAlignment.Center)
            .Columns.Add("Value", 62, HorizontalAlignment.Right)
            .Columns.Add("    Notes", 100, HorizontalAlignment.Left)
            .Refresh()
            ' -----------------------------------------------------------------
            For Each m As Module_InOut In InOutModules
                item = .Items.Add("Module")
                item.SubItems.Add((m.ModuleId + 1).ToString)
                item.SubItems.Add(m.GetName)
                item.SubItems.Add("")
                item.SubItems.Add("")
                item.SubItems.Add(m.ModuleVersion)
                For Each p As Pin In m.Pins
                    If CInt(Val(p.PinName)) < m.Adc24_MasterPinName Then
                        item = .Items.Add(" Pin")
                    ElseIf CInt(Val(p.PinName)) < m.Generic_FirstPinName Then
                        item = .Items.Add(" Adc24")
                    Else
                        item = .Items.Add(" Gener.")
                    End If
                    ' --------------------------------------------------- 
                    item.SubItems.Add(p.PinName)
                    ' ---------------------------------------------------
                    item.SubItems.Add(Pin.PinTypeToString(p.GetPinType))
                    If p.Direction <> Pin.Directions.Unused Then
                        item.SubItems.Add(p.Slot.ToString)
                        item.SubItems.Add("0.0")
                    Else
                        item.SubItems.Add("")
                        item.SubItems.Add("")
                    End If

                    If p.GetPinType = Pin.PinTypes.UNUSED Then
                        item.SubItems.Add("")
                    Else
                        item.SubItems.Add(SlotNames(p.Slot))
                    End If
                Next
            Next
            TheSystem_ListViewResize()
        End With
    End Sub

    ' ------------------------------------------------------------------
    '  Call this function from Form.Resize (not from ListView.resize)
    '  (otherwise the listview scroll produces problems)
    ' ------------------------------------------------------------------
    Friend Sub TheSystem_ListViewResize()
        ' -------------------------------------------- ResizeLastColumn
        With Form1.MyListView1
            If .Columns.Count > 1 Then
                .Columns(.Columns.Count - 1).AutoResize(ColumnHeaderAutoResizeStyle.HeaderSize)
            End If
        End With
    End Sub


    ' ==================================================================================================
    '   LIST COMPONENTS - UPDATE VALUES
    ' ==================================================================================================
    Friend Sub TheSystem_UpdateListedSubtypes()
        Dim str As String
        With Form1.MyListView1
            Dim n As Int32 = 0
            For Each m As Module_InOut In InOutModules
                If m.GetName <> .Items(n).SubItems(2).Text Then .Items(n).SubItems(2).Text = m.GetName
                n += 1
                For Each p As Pin In m.Pins
                    str = Pin.PinTypeToString(p.GetPinType)
                    If str <> .Items(n).SubItems(2).Text Then .Items(n).SubItems(2).Text = str
                    n += 1
                Next
            Next
        End With
    End Sub

    Friend Sub TheSystem_TestModuleError()
        For Each m As Module_InOut In InOutModules
            If Not m.MyDeviceIsActive Or Not m.MyDeviceIsProgrammed Then
                Slots.WriteSlot(CommandSlot, NAN_MasterError)
            End If
        Next
    End Sub

    Friend Sub TheSystem_UpdateListedValues()
        Dim str As String
        With Form1.MyListView1
            Dim n As Int32 = 0
            For Each m As Module_InOut In InOutModules

                ' TODO - DISCONNECTED - Status visualization
                ' ----------------------------------------------------------------- MyDeviceIsActive
                If Not m.MyDeviceIsActive Then
                    If Not m.DisconnectedStatusVisualized Then
                        m.DisconnectedStatusVisualized = True
                        .Items(n).SubItems(5).Text = "Communication lost at " + m.DisconnectionTime.ToString("yyyy/MM/dd HH:mm:ss +fff") + " mS"
                        TheSystem_UpdateListColors(n)
                    End If
                Else
                    If m.DisconnectedStatusVisualized Then
                        m.DisconnectedStatusVisualized = False
                        .Items(n).SubItems(5).Text = m.ModuleVersion
                        TheSystem_UpdateListColors(n)
                        m.ProgrammedStatusVisualized = False
                    End If
                End If

                ' ----------------------------------------------------------------- MyDeviceIsProgrammed
                If Not m.MyDeviceIsProgrammed Then
                    If Not m.ProgrammedStatusVisualized Then
                        m.ProgrammedStatusVisualized = True
                        .Items(n).SubItems(5).Text = "The IotModule has lost the configuration"
                         TheSystem_UpdateListColors(n)
                    End If
                Else
                    If m.ProgrammedStatusVisualized Then
                        m.ProgrammedStatusVisualized = False
                        .Items(n).SubItems(5).Text = m.ModuleVersion
                        TheSystem_UpdateListColors(n)
                        m.DisconnectedStatusVisualized = False
                    End If
                End If

                ' ----------------------------------------------------------------
                n += 1
                For Each p As Pin In m.Pins
                    If p.Direction <> Pin.Directions.Unused Then
                        If Single.IsNaN(p.Value) Then
                            str = NAN_GetName(p.Value)
                        Else
                            str = p.Value.ToString("0.0", GCI)
                        End If
                        If str <> .Items(n).SubItems(4).Text Then .Items(n).SubItems(4).Text = str
                    End If
                    n += 1
                Next
            Next
        End With
    End Sub

    ' TODO - DISCONNECTED - UpdateListColors
    Friend Sub TheSystem_UpdateListColors(ByVal ListLine As Int32)
        With Form1.MyListView1.Items(ListLine)
            If .Text <> "Module" Then Return
            Dim m As Module_InOut = FindModuleByListLine(ListLine)
            If .SubItems(5).Text.StartsWith("Communication") Then
                .BackColor = Color.FromArgb(255, 180, 150)
                .ForeColor = Color.Black
            ElseIf .SubItems(5).Text.StartsWith("The IotModule") Then
                .BackColor = Color.FromArgb(255, 100, 100)
                .ForeColor = Color.Black
            Else
                .BackColor = Color.FromArgb(180, 205, 255)
                .ForeColor = Color.Black
            End If
        End With
    End Sub


    ' ==================================================================================================
    '   CONFIGURATIONS
    ' ==================================================================================================
    Friend ConfigDatabase() As Object

    Private Sub MarkErrorLines()
        Dim config(-1) As String
        Dim nline As Int32 = 0
        '
        DuplicateNames = False
        '
        For Each m As Module_InOut In InOutModules

            If AnotherModuleHasTheSameName(m) Then
                m.StopTimedOperations()
                m.ConfigValid = False
                ConfigurationIsValid = False
                DuplicateNames = True
                Form1.MarkAllLinesWithDuplicateName(m)
                Continue For
            End If

            Dim npin As Int32 = 0
            Dim sa() As String
            '
            If ConfigDatabase.Length > 0 Then
                config = TryCast(ConfigDatabase(m.ConfigId), String())
            End If

            If npin < config.Length Then
                sa = Split(config(npin), ",")
                If sa.Length < 2 OrElse sa(0).Trim <> "Module" OrElse sa(1).Trim <> m.GetName Then
                    MarkErrorLine(m, npin)
                End If
            Else
                MarkErrorLine(m, npin)
            End If
            npin += 1
            nline += 1
            For Each p As Pin In m.Pins
                If npin < config.Length Then
                    sa = Split(config(npin), ",")
                    If sa.Length < 2 OrElse sa(0).Trim <> "Pin" OrElse sa(1).Trim <> Pin.PinTypeToString(p.GetPinType) Then
                        MarkErrorLine(m, npin)
                    End If
                Else
                    MarkErrorLine(m, npin)
                End If
                npin += 1
                nline += 1
            Next
            '
        Next
    End Sub

    Private Sub MarkErrorLine(ByVal m As Module_InOut, ByVal LineNumber As Int32)
        '
    End Sub

    Private Sub AddErrorLine(ByVal config As String(), ByVal npin As Int32, ByVal nline As Int32)
        With Form1.MyListView1
            ' ------------------------------------------------------- prepare the item
            Dim itm As ListViewItem
            Dim s() As String = Split(config(npin), ",")
            If s(0) = "Slave" Then
                itm = New ListViewItem(" " + s(0))
            Else
                itm = New ListViewItem("  " + s(0))
            End If
            itm.SubItems.Add("")
            itm.SubItems.Add(s(1).Trim)
            itm.SubItems.Add("")
            itm.SubItems.Add("")
            itm.SubItems.Add("Not found")
            itm.BackColor = Color.FromArgb(255, 150, 150)
            itm.ForeColor = Color.Black
            ' ------------------------------------------------------- insert the item in position "nline"
            .Items.Insert(nline, itm)
        End With
    End Sub

    Private Function AnotherModuleHasTheSameName(ByVal m1 As Module_InOut) As Boolean
        For Each m2 As Module_InOut In InOutModules
            If m1 Is m2 Then Continue For
            If m1.GetName = m2.GetName Then Return True
        Next
        Return False
    End Function

    Friend Function GetConfigName(ByVal config As String()) As String
        Dim sa() As String
        sa = Split(config(0), ",")
        If sa.Length > 1 Then
            Return sa(1).Trim
        Else
            Return ""
        End If
    End Function

    Friend Function FindConfigByName(ByVal name As String) As Int32
        For i As Int32 = 0 To ConfigDatabase.Length - 1
            If name.ToUpper = GetConfigName(TryCast(ConfigDatabase(i), String())).ToUpper Then
                Return i
            End If
        Next
        Return -1
    End Function

    Friend Sub FillConfigNamesCombo(ByVal cmb As ComboBox)
        For i As Int32 = 0 To ConfigDatabase.Length - 1
            cmb.Items.Add(GetConfigName(TryCast(ConfigDatabase(i), String())))
        Next
    End Sub

    Private Sub AssignConfigurationsToModules()
        '
        ConfigurationIsValid = True
        '
        For i As Int32 = 0 To InOutModules.Count - 1
            InOutModules(i).ConfigValid = True
            ' ----------------------------------------------------------- find configuration by name
            InOutModules(i).ConfigId = FindConfigByName(InOutModules(i).GetName)
            ' ----------------------------------------------------------- if not found add a new configuration
            If InOutModules(i).ConfigId < 0 Then
                InOutModules(i).ConfigId = ConfigDatabase.Length
                Dim config As String() = CreateConfig_AsStringArray(InOutModules(i))
                AddToConfigDatabase(config)
                TheSystem_SaveConfigDatabase()
            End If
        Next
    End Sub

    Private Sub SetConfigurationParamsToDevices()
        Dim sa() As String
        For Each m As Module_InOut In InOutModules
            If m.ConfigId >= ConfigDatabase.Length Then Continue For
            Dim config As String() = TryCast(ConfigDatabase(m.ConfigId), String())
            If config Is Nothing Then Continue For
            Dim nline As Int32 = 0
            If nline >= config.Length Then Continue For
            sa = Split(config(nline), ",")
            nline += 1
            If sa.Length = 3 Then
                m.SetSpeed(CInt(Val(sa(2).Trim)))
            End If

            For Each p As Pin In m.Pins
                ' -----------------------------------------------------------
                Dim ConfLine As String = GetConfigLineByPinName(p.PinName, config)
                If ConfLine = "" Then Continue For
                ' -----------------------------------------------------------
                sa = Split(ConfLine, ",")
                Dim ConfPinType As Pin.PinTypes = Pin.StringToPinType(sa(1))
                ' ----------------------------------------------------------- test if the PinType is in the area
                If Not TestValidPinType(ConfPinType, p.PinName, m) Then
                    ConfPinType = Pin.PinTypes.UNUSED
                End If
                ' ----------------------------------------------------------- SLOW - test the PinType calling module
                'If Not m.TestValidPinType(ConfPinType, p) Then
                '    ConfPinType = Pin.PinTypes.UNUSED
                'End If
                ' -----------------------------------------------------------
                If ConfPinType <> p.GetPinType Then
                    p.SetPinType(ConfPinType)
                End If
                ' -----------------------------------------------------------
                If sa.Length >= 6 Then
                    p.Slot = CInt(Val(sa(2).Trim))
                    If p.Slot > 999 Then p.Slot = 999
                    p.Value_Max = CSng(Val(sa(3).Trim))
                    p.Value_Min = CSng(Val(sa(4).Trim))
                    '
                    Dim nn As Int32 = CInt(Val(sa(5).Trim))
                    p.AdaptiveSpeed = False
                    If nn >= 1000 Then
                        p.AdaptiveSpeed = True
                        nn -= 1000
                    End If
                    If nn > 100 Then nn = 100
                    If nn < 1 Then nn = 1
                    p.ResponseSpeed = nn
                    '
                    Select Case p.GetPinType
                        Case Pin.PinTypes.SERVO
                            If sa.Length >= 8 Then
                                p.MaxTime = CSng(Val(sa(6).Trim))
                                p.MinTime = CSng(Val(sa(7).Trim))
                            End If
                        Case Pin.PinTypes.STEPPER
                            If sa.Length >= 9 Then
                                p.MaxSpeed = CSng(Val(sa(6).Trim))
                                p.MaxAcc = CSng(Val(sa(7).Trim))
                                p.StepsPerMillimeter = CSng(Val(sa(8).Trim))
                            End If
                        Case Pin.PinTypes.PWM
                            If sa.Length >= 8 Then
                                p.PwmFrequency = CSng(Val(sa(6).Trim))
                                p.LogResponse = sa(7).Trim = "True"
                            End If
                        Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                            If sa.Length >= 8 Then
                                p.MinVariation = CSng(Val(sa(6).Trim))
                                p.ProportionalArea = CSng(Val(sa(7).Trim))
                            End If
                        Case Pin.PinTypes.COUNTER, Pin.PinTypes.COUNTER_PU, _
                             Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                            If sa.Length >= 9 Then
                                p.MaxFreq = CSng(Val(sa(6).Trim))
                                p.MinFreq = CSng(Val(sa(7).Trim))
                                p.ConvertToFreq = sa(8).Trim = "True"
                            End If
                            '
                        Case Pin.PinTypes.ADC_24
                            If sa.Length >= 8 Then
                                p.Adc24Sps = CInt(Val(sa(6).Trim))
                                p.Adc24Filter = CInt(Val(sa(7).Trim))
                            End If
                            '
                        Case Pin.PinTypes.ADC_24_CH, _
                             Pin.PinTypes.ADC_24_CH_B
                            If sa.Length >= 9 Then
                                p.Adc24ChType = CInt(Val(sa(6).Trim))
                                p.Adc24ChGain = CInt(Val(sa(7).Trim))
                                p.Adc24ChBias = sa(8).Trim = "True"
                            End If
                            '
                        Case Pin.PinTypes.ENCODER_B, Pin.PinTypes.ENCODER_B_PU
                            ' nothing to do
                    End Select
                End If
            Next
            ' -------------------------------------- re-align all the pins 
            m.SetupModulePins()
            ' -------------------------------------- send special STEPPER parameters
            For Each p As Pin In m.Pins
                If p.GetPinType = Pin.PinTypes.STEPPER Then
                    ' ------------------------------------------------------------------
                    '  STEPPER PIN CONFIGURATION - BY SendBytesToSlave
                    ' ------------------------------------------------------------------
                    p.SendStepperParamsToHardware()
                End If
            Next
            ' -------------------------------------- send ADC_24 configuration
            If m.Pins.Count > 0 AndAlso m.Pins(m.Adc24_MasterPinIndex).GetPinType = Pin.PinTypes.ADC_24 Then
                m.Adc24ConfigAndStart()
            End If
        Next
    End Sub

    Private Function GetConfigLineByPinName(ByVal PinName As String, ByVal config As String()) As String
        For Each line As String In config
            If line.StartsWith("Pin_" + PinName + ",") Then
                Return line
            End If
        Next
        Return ""
    End Function

    Private Function TestValidPinType(ByVal PinType As Pin.PinTypes, ByVal PinName As String, ByVal m As Module_InOut) As Boolean
        Dim PinNameInteger As Int32 = CInt(Val(PinName))
        Select Case PinType
            Case Pin.PinTypes.ADC_24, Pin.PinTypes.ADC_24_CH, Pin.PinTypes.ADC_24_CH_B
                If PinNameInteger < m.Adc24_MasterPinName Or PinNameInteger > m.Adc24_LastPinName Then
                    Return False
                End If
            Case Pin.PinTypes.GEN_IN_8, Pin.PinTypes.GEN_IN_16, _
                 Pin.PinTypes.GEN_IN_24, Pin.PinTypes.GEN_IN_FLOAT, _
                 Pin.PinTypes.GEN_OUT_8, Pin.PinTypes.GEN_OUT_16, _
                 Pin.PinTypes.GEN_OUT_24, Pin.PinTypes.GEN_OUT_FLOAT
                If PinNameInteger >= m.Adc24_MasterPinName And PinNameInteger <= m.Adc24_LastPinName Then
                    Return False
                End If
            Case Else
                If PinNameInteger >= m.Adc24_MasterPinName Then
                    Return False
                End If
        End Select
        Return True
    End Function

    Private Function CreateConfig_AsStringArray(ByVal m As Module_InOut) As String()
        Dim i As Int32
        Dim config(0) As String
        config(0) = "Module"
        config(0) &= "," & m.GetName
        config(0) &= "," & m.CommSpeed.ToString
        For Each p As Pin In m.Pins
            i = config.Length
            ' -----------------------------------------------------------
            If p.GetPinType = Pin.PinTypes.UNUSED Then Continue For
            ReDim Preserve config(i)
            config(i) = "Pin_" + p.PinName
            config(i) &= "," & Pin.PinTypeToString(p.GetPinType)
            config(i) &= "," & p.Slot.ToString
            config(i) &= "," & p.Value_Max.ToString(GCI)
            config(i) &= "," & p.Value_Min.ToString(GCI)
            Dim nn As Int32 = p.ResponseSpeed
            If p.AdaptiveSpeed Then nn += 1000
            config(i) &= "," & nn.ToString
            Select Case p.GetPinType
                Case Pin.PinTypes.SERVO
                    config(i) &= "," & p.MaxTime.ToString
                    config(i) &= "," & p.MinTime.ToString
                Case Pin.PinTypes.STEPPER
                    config(i) &= "," & p.MaxSpeed.ToString(GCI)
                    config(i) &= "," & p.MaxAcc.ToString(GCI)
                    config(i) &= "," & p.StepsPerMillimeter.ToString(GCI)
                Case Pin.PinTypes.PWM
                    config(i) &= "," & p.PwmFrequency.ToString
                    config(i) &= "," & p.LogResponse.ToString
                Case Pin.PinTypes.CAP_8, Pin.PinTypes.CAP_16
                    config(i) &= "," & p.MinVariation.ToString
                    config(i) &= "," & p.ProportionalArea.ToString
                Case Pin.PinTypes.COUNTER, Pin.PinTypes.COUNTER_PU, _
                     Pin.PinTypes.PERIOD, Pin.PinTypes.PERIOD_PU
                    config(i) &= "," & p.MaxFreq.ToString
                    config(i) &= "," & p.MinFreq.ToString
                    config(i) &= "," & p.ConvertToFreq.ToString
                Case Pin.PinTypes.ADC_24
                    config(i) &= "," & p.Adc24Sps.ToString
                    config(i) &= "," & p.Adc24Filter.ToString
                Case Pin.PinTypes.ADC_24_CH, Pin.PinTypes.ADC_24_CH_B
                    config(i) &= "," & p.Adc24ChType.ToString
                    config(i) &= "," & p.Adc24ChGain.ToString
                    config(i) &= "," & p.Adc24ChBias.ToString
                Case Pin.PinTypes.ENCODER_B, Pin.PinTypes.ENCODER_B_PU
                    ' nothing to do
            End Select
        Next
        Return config
    End Function

    Friend Sub TheSystem_ValidateConfiguration()
        For Each m As Module_InOut In InOutModules
            m.ConfigValid = True
        Next
        ConfigurationIsValid = True
        TheSystem_ListComponents()
        Form1.ListView_SetAllLineColors()
        Form1.ToolStripButton_Validate.Enabled = Not ConfigurationIsValid And Not DuplicateNames
        Form1.EnableCalibrateButton(TheSystem_CalibrationNeeded)
    End Sub

    Friend Sub AddToConfigDatabase(ByVal config As String())
        If config.Length > 0 Then
            Dim count As Int32 = ConfigDatabase.Length
            ReDim Preserve ConfigDatabase(count)
            ConfigDatabase(count) = config
        End If
    End Sub

    Friend Sub TheSystem_UpdateAndSaveConfigDatabase()
        TheSystem_UpdateConfigDatabase()
        TheSystem_SaveConfigDatabase()
    End Sub

    Friend Sub TheSystem_UpdateConfigDatabase()
        If ConfigurationIsValid Then
            For Each m As Module_InOut In InOutModules
                If m.ConfigValid Then
                    If m.ConfigId <= ConfigDatabase.Length And ConfigDatabase.Length > 0 Then
                        ConfigDatabase(m.ConfigId) = CreateConfig_AsStringArray(m)
                    Else
                        m.StopTimedOperations()
                        m.ConfigValid = False
                        ConfigurationIsValid = False
                    End If
                End If
            Next
        End If
    End Sub

    Friend Sub TheSystem_SaveConfigDatabase()
        If ConfigurationIsValid Then
            Save_ConfigDatabase()
        Else
            'MsgBox("Invalid configuration - ""Config Database"" not saved.", MsgBoxStyle.Information)
        End If
    End Sub


End Module
